# into PCM_MULAW format audio files, in order to be able to play prerecorded messages

import argparse
import glob
import os
import pathlib
import subprocess  # Needed to sent commands to the linux terminal

# Anton imports
from utils import get_logger, get_line_number, check_wav_file_format
import shared_variables
import constants
import time
from datetime import datetime
import sys

logger = get_logger('AudioFilesProcessor')


class AudioFilesProcessor:
    """Class to convert audio files from mp3 format into PCM_MULAW format wav file."""

    def __init__(self, raw_dir=constants.FILE_PATH, processed_dir=constants.PROCESSED_FILES_PATH,):
        self.__raw_files_dir = raw_dir
        self.__raw_audio_files = glob.glob(os.path.join(raw_dir, "*.mp3"))
        self.__processed_dir = processed_dir
        self.__format = "pcm_mulaw"  # pcm_s16le, pcm_mulaw
        self.__frequency_8 = 8000
        self.__frequency_24 = 24000
        pathlib.Path(self.__processed_dir).mkdir(parents=True, exist_ok=True)

    def _process_audio_files(self, old_file="", new_file=""):
        """
        Processes all the mp3 audio files located in raw_dir using ffmpeg.
        Note: Only called at system start-up
        """
        for raw_audio_file in self.__raw_audio_files:
            try:
                if self._check_for_raw_file_updates(raw_audio_file) is constants.YES:
                    logger.info(f"Raw file to be processed: {raw_audio_file}")

                    update_dict = self._processFile(raw_audio_file)
                    shared_variables.audio_files.update_timestamp(**update_dict)
            except KeyError as e:
                logger.warning(f"Missing key: {e}")
                logger.warning("Make sure that all the physical files have equivalent in the backup.json file")
            except Exception as e:
                logger.warning(f"Caught following exception: {e}")

        if len(self.__raw_audio_files) == 0:
            logger.warning("No audio files found!")

    def _delete_processed_files(self):
        """
        Deletes all the .wav files from the processed audio files directory.
        Note: Call this method to clear out all older files from the directory.
        """
        if shared_variables.delete_processed_files == constants.YES:
            processed_audio_files = glob.glob(os.path.join(self.__processed_dir, "*.wav"))
            if len(processed_audio_files) != 0:
                logger.info(f"Deleting processed audio files")
                for audio_file in processed_audio_files:
                    os.remove(audio_file)
                    logger.info(f"{audio_file} deleted!")
            else:
                logger.info("No files to process. Folder empty!")
        else:
            logger.info("Delete files flag set to NO -> existing files will not be deleted")

    def _check_for_raw_file_updates(self, file):
        """
            Checks if the current WAV files need to be encoded again, in the case where there is a
            newer version of their corresponding raw MP3 file. This should be just an exceptional case
            as under normal circumstances the file is encoded on file upload.
        Args:
            file: the file with its filepath to be checked
        Returns:
            constants.YES  -> when there is an updated raw file
            constants.NO -> no update is required
        """
        command = [
            'stat',
            '-c',
            '%Z',
            file
        ]

        process = subprocess.Popen(
            command,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        try:
            outs, errs = process.communicate(timeout=5)

            timestamp_epoch = int(outs)

            # Get file dictionary
            old_file_dict = shared_variables.audio_files.get_timestamp_from_name(file)
            utc_time = datetime.strptime(old_file_dict["timestamp"], "%Y-%m-%d %H:%M:%S")
            old_file_epoch = (utc_time - datetime(1970, 1, 1)).total_seconds()

            logger.debug("output: %s", str(outs))
            logger.debug(f"error :{errs}")
            if old_file_epoch < timestamp_epoch:
                logger.warning("Found file discrepancy. Encoding update is needed")
                return constants.YES
            else:
                return constants.NO

        except subprocess.TimeoutExpired:
            process.kill()
            outs, errs = process.communicate()
            logger.warning(
                f"Timed out, killing check of {file} : errs :{errs}"
            )
            return constants.NO

    def _check_processed_audio_files(self):
        """
        Checks if audio files in mp3 format have processed counterpart understandable by
        Paging Server's ringtone/fileplay service.
        """
        # Get the actual physical files from the device again. Important after adding/removing files
        self.__raw_audio_files = glob.glob(os.path.join(self.__raw_files_dir, "*.mp3"))

        # Get the actual physical files from the device
        processed_audio_files = glob.glob(os.path.join(self.__processed_dir, "*.wav"))
        processed_station_names = set(
            map(lambda x: os.path.splitext(os.path.basename(x))[0], processed_audio_files)
        )
        unprocessed_station_names = set(
            map(lambda x: os.path.splitext(os.path.basename(x))[0], self.__raw_audio_files)
        )

        if processed_station_names != unprocessed_station_names:
            logger.warning(
                f"Mismatch found in processed and non processed audio files"
            )
            logger.warning(
                f"MP3 audio file for stations :{unprocessed_station_names}"
            )
            logger.warning(
                f"WAV audio file for stations :{processed_station_names}"
            )

    def _create_timestamp(self) -> str:
        timestamp_date = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        return str(timestamp_date)

    def _processFile(self, raw_audio_file) -> str:
        """
        Encodes an *.mp3 file to a *.wav file which can be played by the device
        Args:
            raw_audio_file: MP3 audio file
        Return:
            timestamp: Time at which the file was processed
        """
        logger.info(f"Processing {raw_audio_file}")
        audio_file_name = os.path.splitext(os.path.basename(raw_audio_file))[0]
        file_dict = {
            "file_id": 0,
            "timestamp": "1970-01-01 00:00:00"
        }

        command = [
            "ffmpeg",
            "-y",  # to overwrite the output file if exists.
            "-i",
            raw_audio_file,
            "-acodec",
            f"{self.__format}",
            "-ar",
            f"{self.__frequency_8}",
            "-ac",
            "1",
            os.path.join(self.__processed_dir, f"{audio_file_name}.wav"),
        ]
        process = subprocess.Popen(
            command,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        try:
            outs, errs = process.communicate(timeout=8)  # Longer timeout so larger files can be processed

            # Create a timestamp after the file has been processed
            new_timestamp = self._create_timestamp()
            file_dict = shared_variables.audio_files.get_timestamp_from_name(raw_audio_file)
            file_dict["timestamp"] = new_timestamp

            logger.debug(f"output: {outs}")
            logger.debug(f"error :{errs}")
        except subprocess.TimeoutExpired:
            process.kill()
            outs, errs = process.communicate()
            logger.warning(
                f"Timed out, killing processing of {audio_file_name}.mp3 : errs :{errs}"
            )
        return file_dict

    def _purgeFile(self, file_to_delete, erase_original_file):
        """
        Removes both the original(when the flag is set) and processed file
        Args:
            file_to_delete: Audio file name stored during file adding
            erase_original_file: Flag to determine whether the original file should also be deleted
        """
        wav_file = check_wav_file_format(file_to_delete)
        wav_file = os.path.join(self.__processed_dir, wav_file)
        mp3_file = os.path.join(self.__raw_files_dir, file_to_delete)
        # Get the actual physical files from the device
        processed_audio_files = glob.glob(os.path.join(self.__processed_dir, "*.wav"))

        if len(processed_audio_files) != 0:
            if wav_file in processed_audio_files:
                logger.info(f"Deleting processed audio file...")
                os.remove(wav_file)
                logger.info(f"{wav_file} deleted!!!")
            else:
                logger.error("The given %s was not found in processed files! [%s]", wav_file, get_line_number())

            if mp3_file in self.__raw_audio_files and erase_original_file:
                logger.info(f"Deleting original audio file...")
                os.remove(mp3_file)
                logger.info(f"{mp3_file} deleted!!!")
            else:
                if not erase_original_file:
                    logger.info("erase_original_file flag not set")
                else:
                    logger.error("The given %s was not found in raw files! [%s]", mp3_file, get_line_number())
        else:
            logger.info("No files to process. Folder empty!")

    def run(self):
        """
        Deletes older .wav files from processed directory and creates processed files from mp3
        audio files located in raw directory.
        """
        # self._delete_processed_files()
        self._process_audio_files()
        self._check_processed_audio_files()

    def addNewFile(self, raw_audio_file):
        time.sleep(1)  # Slow down the file adding, to give time for the OS to update the filepaths
        file_dict = self._processFile(raw_audio_file)
        self._check_processed_audio_files()
        return file_dict["timestamp"]

    def removeFile(self, file_to_delete, erase_original_file):
        self._purgeFile(file_to_delete, erase_original_file)
        self._check_processed_audio_files()


def main():
    # files_path_raw = "/home/anton/Music"
    # files_path_processed = files_path_raw + "/processed"

    parser = argparse.ArgumentParser(
        description="Processes mp3 audio files to format understood by Paging Server's ringtone service"
    )
    parser.add_argument(
        "--raw_dir",
        help="Path to directory containing mp3 audio files",
        type=str,
        default=constants.FILE_PATH,
    )
    parser.add_argument(
        "--proc_dir",
        help="Path to directory containing mp3 audio files",
        type=str,
        default=constants.PROCESSED_FILES_PATH,
    )

    args = parser.parse_args()
    audio_files_processor = AudioFilesProcessor(args.raw_dir, args.proc_dir)
    audio_files_processor.run()
    logger.info("Anton, the files have been successfully converted. Check %s directory", constants.PROCESSED_FILES_PATH)


if __name__ == "__main__":
    main()

