import audioop
import json
import threading
import time
from http.server import BaseHTTPRequestHandler, HTTPServer
from secrets import token_urlsafe
from threading import Thread, enumerate
from typing import Tuple

import shared_variables
from utils import get_logger, check_wav_file_format

# Anton imports
import os
import sys
import constants
import traceback
from multiprocessing import Process
from queue import Queue

logger = get_logger("HTTP_Server_backend")


class PagingHttpRequestHandler(BaseHTTPRequestHandler):
    def _set_response(self, code: int = 200):
        self.send_response(code)
        self.send_header("Content-type", "text/html")
        self.end_headers()

    def do_GET(self):
        # logger.warning("Received a GET request. path = %s", self.path)
        if self.path == "/clients":
            self._return_active_clients()
        elif self.path == "/groups":
            self._return_groups()
        elif self.path == "/audio_files":
            self._return_audio_files()
        elif self.path == "/users":
            self._return_users()
        else:
            self._set_response(404)

    def do_PUT(self):
        # logger.warning("Received a PUT request. path = %s", self.path)
        if self.path == "/paging":
            self._page_audio()
        elif self.path == "/recording":
            self._record_audio()
        elif self.path == "/playfile":
            self._get_playfile_state()
        else:
            self._set_response(404)

    def do_POST(self):
        """
        Post request for starting/stopping paging request.
        """
        # logger.warning("Received a POST request. path = %s", self.path)
        if self.path == "/paging":
            self._start_stop_page_message()
        elif self.path == "/recording":
            self._start_stop_page_message()
        elif self.path == "/groups":
            self._create_group()
        elif self.path == "/users":
            self._create_user()
        elif self.path == "/audio_files":
            self._create_audio_file()
        elif self.path == "/playfile":
            self._start_stop_playing_file()
        else:
            self._set_response(404)

    def do_DELETE(self):
        if self.path == "/groups":
            self._delete_group()
        elif self.path == "/users":
            self._delete_user()
        elif self.path == "/audio_files":
            self._delete_audio_file()
        else:
            self._set_response(404)

    def do_UPDATE(self):
        if self.path == "/groups":
            self._update_group()
        elif self.path == "/audio_files":
            self._update_audio_file()
        elif self.path == "/users":
            self._update_user()
        else:
            self._set_response(404)

    def _get_playfile_state(self):
        """
            While a file is being played, the front end can get the status by sending a /playfile_state request
        """
        content_length = int(self.headers["Content-Length"])
        message = self.rfile.read(content_length)
        message = json.loads(message)
        if self._check_user(message["token"]):
            if shared_variables.playfile_state == constants.BUSY:
                self._set_response(206)
                # logger.info(">>> Audio file is currently being played <<<")
            elif shared_variables.playfile_state == constants.IDLE:
                logger.info(">>> No audio file is currently being played <<<")
                self._set_response(200)
            else:
                logger.warning(">>> Unknown playfile state <<<")
                self._set_response(500)
        else:
            logger.warning(f"Internal server error, {message}")
            logger.warning("Unable to authenticate user")
            # The paging state is not ended correctly
            self._set_response(500)

    # Play a message that has been stored in the memory of the device
    def _start_stop_playing_file(self):
        shared_variables.downloading_recorded_message = constants.NO
        playfile_group_thread = None
        playfile_playing_thread = None
        content_length = int(self.headers["Content-Length"])
        message = self.rfile.read(content_length)
        # logger.info("Inside _start_stop_playing_file().")
        # logger.info("Message: %s", message)
        message = json.loads(message)

        if self._check_user(message["token"]):
            # This is the correct order of the messages when playfile should be STOPPED
            if message["type"] == 8 and shared_variables.last_message_type == 3:
                if message["groups"] == "":
                    logger.warning("No groups selected")
                else:
                    if playfile_group_thread is not None:  # Make sure no zombie process is left
                        logger.warning("$$$$$$ REMOVING EXISTING THREAD playfile_group_thread")
                        playfile_group_thread.join()
                        playfile_group_thread = None
                    if playfile_playing_thread is not None:  # Make sure no zombie process is left
                        logger.warning("$$$$$$ REMOVING EXISTING THREAD playfile_playing_thread")
                        playfile_playing_thread.join()
                        playfile_playing_thread = None

                    clients = shared_variables.groups.get_clients(message["groups"])
                    shared_variables.pmaster.send_end_group_request(clients)

                shared_variables.last_client = None
                shared_variables.last_message_type = 8
                shared_variables.active_user = None
                shared_variables.playfile_state = constants.IDLE
                shared_variables.kill_playfile_thread = constants.YES

                # logger.warning("Running threads left")
                # for thread in enumerate():
                #     logger.warning(thread.name)
                logger.info("STOP playing file")
                self._set_response(204)

            # This is the correct order of the messages when playfile should be STARTED
            elif message["type"] == 3 and shared_variables.last_message_type == 8:

                if message["groups"] == "":
                    logger.warning("No groups selected")
                else:
                    clients = shared_variables.groups.get_clients(message["groups"])
                    shared_variables.last_client = clients
                    playfile_group_thread = Thread(
                        target=shared_variables.pmaster.send_page_group_request,
                        kwargs={"clients": clients, "priority": 0},
                    )
                    playfile_group_thread.daemon = True
                    playfile_group_thread.start()
                    # logger.info("playfile_group_thread started")

                shared_variables.last_message_type = 3
                logger.info("START playing file")

                # The file starts playing from a separate thread
                try:
                    playfile_playing_thread = Thread(
                        target=self._play_message,
                        # kwargs={"file_id": message["file_ID"]},
                        args=(message["file_ID"],)
                    )
                    playfile_playing_thread.daemon = True
                    playfile_playing_thread.start()
                    # logger.info("New playfile_playing_thread started as %s",
                    #             playfile_playing_thread.getName())
                except TypeError:
                    logger.warning(f"Oops: {sys.exc_info()[0]} occurred")
                except OSError as error:
                    logger.warning("While in _play_message() the following error occurred:")
                    logger.warning(error)

                # logger.warning("Current running threads")
                # for thread in enumerate():
                #     logger.warning(thread.name)

                # logger.warning(f"Current playfile_group_thread value: {playfile_group_thread}")
                # logger.warning(f"Current playfile_playing_thread value: {playfile_playing_thread}")
                # logger.info("About to start the playing")
                shared_variables.first_fast_packets = 0
                self._set_response(204)

            # In case the last message was not ended correctly
            elif message["type"] == 3:
                logger.warning("Last message not ended properly. Ending it first!")
                if shared_variables.last_client:
                    if playfile_group_thread is not None:
                        logger.warning("$$$$$$ REMOVING EXISTING THREAD playfile_playing_thread. Not proper ending")
                        playfile_group_thread.join()
                        playfile_group_thread = None
                    if playfile_playing_thread is not None:
                        logger.warning("$$$$$$ REMOVING EXISTING THREAD playfile_playing_thread. Not proper ending")
                        playfile_playing_thread.join()
                        playfile_playing_thread = None

                    shared_variables.pmaster.send_end_group_request(
                        shared_variables.last_client
                    )
                if message["groups"] == "":
                    shared_variables.last_client = None
                    logger.warning("No group selected")
                else:
                    clients = shared_variables.groups.get_clients(message["groups"])
                    shared_variables.last_client = clients
                    playfile_group_thread = Thread(
                        target=shared_variables.pmaster.send_page_group_request,
                        kwargs={"clients": clients, "priority": 0},
                    )
                    playfile_group_thread.daemon = True
                    playfile_group_thread.start()
                    logger.info("playfile_group_thread started.")

                shared_variables.last_message_type = 3
                logger.info("START playing file")

                try:
                    playfile_playing_thread = Thread(
                        target=self._play_message,
                        args=(message["file_ID"],)
                    )
                    playfile_playing_thread.daemon = True
                    playfile_playing_thread.start()
                    logger.info("playfile_playing_thread started")
                except TypeError:
                    logger.warning(f"Oops: {sys.exc_info()[0]} occured")
                except OSError as error:
                    logger.warning("While in _play_message() the following error occured:")
                    logger.warning(error)

                shared_variables.first_fast_packets = 0
                self._set_response(204)
                shared_variables.playfile_state = constants.IDLE

            else:
                logger.warning(f"Internal server error, {message}")
                # The playgile state is not ended correctly
                self._set_response(500)
        else:
            error_msg = f"Another user is playing: {shared_variables.active_user}"
            new_user = shared_variables.users.get_user_from_token(message["token"])
            logger.info(f"Received request from {new_user}")
            logger.warning(error_msg)
            self._set_response(409)
            self.send_header("Content-type", "application/json")
            self.end_headers()
            self.wfile.write(json.dumps(error_msg).encode())

    def _play_message(self, file_id=""):

        shared_variables.downloading_recorded_message = constants.NO
        shared_variables.first_fast_packets = 0
        shared_variables.kill_playfile_thread = constants.NO
        local_playback_process = None
        audio_file = None

        file_id = check_wav_file_format(file_id)

        files = []  # Store all the files from the given directory
        # if os.listdir(constants.PROCESSED_FILES_PATH):
        #     logger.info("Available files in the given directory (%s): ", constants.PROCESSED_FILES_PATH)
        for file in os.listdir(constants.PROCESSED_FILES_PATH):
            files.append(file)
            # logger.info("\t" + file)

        if file_id not in files:
            logger.warning("The requested file ID: %s does not exist", file_id)
            shared_variables.playfile_state = constants.IDLE  #
            shared_variables.kill_playfile_thread = constants.YES
        else:
            logger.info("File [%s] found in existing files", file_id)

        if shared_variables.kill_playfile_thread is not constants.YES:

            file_to_open = os.path.join(constants.PROCESSED_FILES_PATH, file_id)
            try:
                fp = open(file_to_open, 'rb')
                audio_file = fp.read()
                fp.close()
                # convert the data to 16-bit signed
                audio_file = audioop.ulaw2lin(audio_file, 2)
            except Exception as ex:
                logger.warning("Unable to read the given %s into bytes: %s", file_to_open, ex)
                shared_variables.kill_playfile_thread = constants.YES
            del file_to_open  # Variable not needed anymore

            # Local playback
            try:
                local_playback_process = Process(
                    target=shared_variables.local_alsa_player.write,
                    args=(audio_file,)
                )
                local_playback_process.daemon = True
                local_playback_process.start()
                # logger.warning("New local_playback_process started as: %s!", local_playback_process.name)
            except TypeError:
                logger.warning(f"Oops: {sys.exc_info()[0]} occurred")
            except OSError as error:
                logger.warning("While in shared_variables.local_alsa_player.write() the following error occurred:")
                logger.warning("Check audio device: %s", shared_variables.config["audio_device"]["playback"])
                logger.warning(error)
        else:
            logger.info("Unable to proceed with execution. shared_variables.kill_playfile_thread is set to: %s !!!",
                        shared_variables.kill_playfile_thread)

        # sent to groups
        if (shared_variables.pmaster.state == 130) and \
                audio_file is not None and \
                shared_variables.kill_playfile_thread is not constants.YES:  # MSG 130, send custom command

            data_block_size = 1280
            # delay = float("{0:.4f}".format(data_block_size / 2 / 8 / 1000))
            delay = 0.079

            if shared_variables.config["sample_rate"] == 8000:
                data_block_size = 1280
                # data_block_size = 320
                delay = float("{0:.4f}".format(data_block_size / 2 / 8 / 1000)) - 0.001
                # delay = 0.080
            elif shared_variables.config["sample_rate"] == 24000:
                data_block_size = 1280
                delay = float("{0:.4f}".format(data_block_size / 2 / shared_variables.config["sample_rate"])) - 0.001
                # delay = 0.040
            else:
                logger.warning("Sample rate not supported")

            leftover = len(audio_file) % data_block_size
            count = 0
            # time.sleep(0.3)  # TODO: Delay for debugging of IPAC. Remove when IPAC is fixed

            if shared_variables.config["audio_format"] == "ulaw":
                shared_variables.playfile_state = constants.BUSY
                logger.info("Using %s audio format", shared_variables.config["audio_format"])
                logger.info(f"Sending {data_block_size/2} byte packages with {delay}s delay")

                for i in range(len(audio_file) // data_block_size):
                    # Needed in order to be able to properly kill the thread prematurely
                    if shared_variables.kill_playfile_thread is not constants.YES:
                        if shared_variables.first_fast_packets < 3:
                            time.sleep(0.002)
                            time.sleep(delay)  # data_block_size//2//8 ms
                            shared_variables.first_fast_packets += 1
                        else:
                            time.sleep(delay)  # data_block_size//2//8 ms

                        shared_variables.pmaster.audio_socket.sendto(
                            audioop.lin2ulaw(audio_file[i * data_block_size: (i + 1) * data_block_size], 2),
                            (
                                shared_variables.pmaster.multicast_group,
                                shared_variables.pmaster.audio_port,
                            ),
                        )
                        count = count + data_block_size
                    else:
                        logger.warning("File playing stopped prematurely.")
                        break

                if leftover != 0:
                    shared_variables.pmaster.audio_socket.sendto(
                        audioop.lin2ulaw(audio_file[len(audio_file) - leftover:], 2),
                        (
                            shared_variables.pmaster.multicast_group,
                            shared_variables.pmaster.audio_port,
                        ),
                    )
                    count = count + leftover
                logger.info(f"Sent this many bytes: {count/2}, out of {len(audio_file)/2}")

            elif shared_variables.config["audio_format"] == "alaw":
                shared_variables.playfile_state = constants.BUSY
                logger.info("Using %s audio format", shared_variables.config["audio_format"])

                for i in range(len(audio_file) // data_block_size):
                    # Needed in order to be able to properly stop playing prematurely
                    if shared_variables.kill_playfile_thread is not constants.YES:
                        shared_variables.pmaster.audio_socket.sendto(
                            audioop.lin2alaw(audio_file[i * data_block_size: (i + 1) * data_block_size], 2),
                            (
                                shared_variables.pmaster.multicast_group,
                                shared_variables.pmaster.audio_port,
                            ),
                        )
                        count = count + data_block_size
                        if shared_variables.first_fast_packets < constants.PACKETS_NUMBER:
                            # time.sleep(0.002)
                            time.sleep(delay)  # data_block_size//2//8 ms
                            shared_variables.first_fast_packets = shared_variables.first_fast_packets + 1
                        else:
                            time.sleep(delay)  # data_block_size//2//8 ms
                    else:
                        logger.warning("File playing stopped prematurely.")
                        break

                if leftover != 0:
                    shared_variables.pmaster.audio_socket.sendto(
                        audioop.lin2alaw(audio_file[len(audio_file) - leftover:], 2),
                        (
                            shared_variables.pmaster.multicast_group,
                            shared_variables.pmaster.audio_port,
                        ),
                    )
                    count = count + leftover
                logger.info(f"Sent this many bytes: {count/2}, out of {len(audio_file)/2}")

            else:
                logger.warning(
                    f"Audio format not recognized : {shared_variables.config['audio_format']}"
                )
                self._set_response(400)

            # Kill the local playback process. Important for the case when the user stops the playback earlier
            if local_playback_process is not None:
                local_playback_process.terminate()
                local_playback_process.join()
                while local_playback_process.is_alive():
                    logger.info("Still alive local_playback_process")
                    time.sleep(1)
                local_playback_process = None

            shared_variables.playfile_state = constants.IDLE
            logger.info("Successfully finished playing the file")

        else:
            # The request could not be processed because of conflict in the current state of the resource,
            self._set_response(409)

    def _record_audio(self):
        local_playback_process = None

        try:
            content_length = int(
                self.headers["Content-Length"]
            )  # <--- Gets the size of data
        except Exception as e:
            logger.error(f"Error while getting content: {e}")
            return self._set_response(409)

        try:
            message = self.rfile.read(content_length)
            token, audio = self._extract_token_and_audio(message)
        except Exception as e:
            logger.error(f"Error while reading the content: {e}")
            return self._set_response(409)

        shared_variables.downloading_recorded_message = constants.YES

        try:
            if token == shared_variables.users.secret(shared_variables.active_user):
                # local playback
                # shared_variables.local_alsa_player.write(audio)

                # Local playback
                try:
                    local_playback_process = Process(
                        target=shared_variables.local_alsa_player.write,
                        args=(audio,)
                    )
                    local_playback_process.daemon = True
                    local_playback_process.start()
                    # logger.warning("New local_playback_process started as: %s!", local_playback_process.name)
                except TypeError:
                    logger.warning(f"Oops: {sys.exc_info()[0]} occurred")
                except OSError as error:
                    logger.warning("While in shared_variables.local_alsa_player.write() the following error occurred:")
                    logger.warning(error)
                    logger.warning("Check audio device: %s", shared_variables.config["audio_device"]["playback"])

                # sent to groups
                if shared_variables.pmaster.state == 130:
                    data_block_size = 480
                    delay = 0.03
                    logger.info("Sample rate: %s", shared_variables.config["sample_rate"])
                    if shared_variables.config["sample_rate"] == 8000:
                        data_block_size = 912
                        delay = float("{0:.4f}".format(data_block_size / 2 / shared_variables.config["sample_rate"]))
                        # delay = 0.02
                    elif shared_variables.config["sample_rate"] == 24000:
                        data_block_size = 964
                        delay = float("{0:.4f}".format(data_block_size / 2 / shared_variables.config["sample_rate"]))
                        # delay = 0.02
                    else:
                        logger.warning("Sample rate not supported")

                    # First record the message on the disk
                    with open(shared_variables.record_service.current_file, "ab") as message_file:
                        logger.info(f"Writing audio to: {shared_variables.record_service.current_file}")
                        message_file.write(audio)  # PCM 16bit

                    # Notify that the file has been downloaded
                    shared_variables.downloading_recorded_message = constants.NO
                    time.sleep(0.15)  # Give some time to read the flag
                    # Then read from the recorded file
                    with open(shared_variables.record_service.current_file, "rb") as message_file:
                        audio_file = message_file.read()  # PCM 16bit

                        logger.info("Reading from recorded file...")
                        # logger.info(f"Current fast packets value: {shared_variables.first_fast_packets}")
                        self._encode_send_audio(audio_file, shared_variables.config["audio_format"],
                                                data_block_size, delay)
                else:
                    logger.warning("Current paging master state: %s", shared_variables.pmaster.state)

                # Kill the local playback process. Important for the case when the user stops the playback earlier
                if local_playback_process is not None:
                    local_playback_process.terminate()
                    local_playback_process.join()
                    while local_playback_process.is_alive():
                        logger.info("Still alive local_playback_process")
                        time.sleep(1)
                    local_playback_process = None

                self._set_response(204)  # Notify front-end we are done
            else:
                logger.error(f"Could not authenticate user with ID: {token}")
                self._set_response(409)

        except KeyError as e:
            # The request could not be processed because of conflict in the current state of the resource,
            if shared_variables.active_user is None:
                logger.error(f"Key error. The current user is: {e}")
            else:
                logger.error(f"Key error. Message: {e}")
                traceback.print_exc()
            self._set_response(409)

    def _page_audio(self):
        shared_variables.downloading_recorded_message = constants.NO
        content_length = int(
            self.headers["Content-Length"]
        )  # <--- Gets the size of data
        message = self.rfile.read(content_length)
        token, audio = self._extract_token_and_audio(message)

        try:
            if token == shared_variables.users.secret(shared_variables.active_user):
                # local playback
                # shared_variables.local_alsa_player.write(audio)  # PCM 16

                # send to groups
                if shared_variables.pmaster.state == 130:
                    shared_variables.paging_message_q.put(audio)  # Add data in the que
                    shared_variables.paging_message_q_buff_msg += 1
                    # logger.info(f"Added to queue {len(audio)}")

                    # Start the queue reading thread as soon as there is data in the queue. Always 16Kb in the buff.
                    # This way prevent problems with very short button clicks.
                    if shared_variables.paging_message_q_buff_msg == 1:
                        shared_variables.paging_message_q_thread.start()  # This thread is only reading from the queue
                    time.sleep(0.01)  # Give chance to other processes to work [10ms]
                self._set_response(204)

        except KeyError as e:
            # The request could not be processed because of conflict in the current state of the resource,
            if shared_variables.active_user is None:
                logger.error(f"Key error. The current user is: {e}")
            else:
                logger.error(f"Key error. Message: {e}")
                traceback.print_exc()
            self._set_response(409)

    def _encode_send_audio_queue(self, m_queue, audio_format="", block_size=1920, delay=0.001):
        """
        Only use this function for live paging
        Reads PCM 16 audio buffers from a message queue, encodes them to either uLaw or aLaw and sends them over
        through the socket to the selected clients.
        Args:
            m_queue: Message queue holding binary IO audio buffers
            audio_format: uLaw or aLaw allowed
            block_size: Packet size to be encoded and sent at a time
            delay: Delay in [s] between the different packets

        Returns: Void
        """
        # Run until the paging_state is changed to IDLE
        while shared_variables.paging_state:
            audio_bin_buff = m_queue.get()
            q_size = m_queue.qsize()

            count = 0
            length = len(audio_bin_buff)
            leftover = length % block_size

            # logger.info(f"Block size: {block_size}, delay: {delay}, q_len: {q_size}")
            # logger.info(f"Value of first fast packets variable: {shared_variables.first_fast_packets}")

            if audio_format == "ulaw":
                for i in range(length // block_size):
                    # elapsed_time = time.perf_counter() - self._start_time  // i
                    shared_variables.pmaster.audio_socket.sendto(
                        audioop.lin2ulaw(audio_bin_buff[i * block_size: (i + 1) * block_size], 2),
                        (
                            shared_variables.pmaster.multicast_group,
                            shared_variables.pmaster.audio_port,
                        ),
                    )
                    count = count + block_size

                    # Burst the first PACKETS_NUMBER packages to accumulate a buffer to read from,
                    # then go with equal intervals
                    if shared_variables.first_fast_packets < constants.PACKETS_NUMBER:
                        time.sleep(0.004)
                        shared_variables.first_fast_packets = shared_variables.first_fast_packets + 1
                    else:
                        time.sleep(delay)

                # If there is data left over from the buffer, send it at the end without delay
                if leftover != 0:
                    shared_variables.pmaster.audio_socket.sendto(
                        audioop.lin2ulaw(audio_bin_buff[length - leftover:], 2),
                        (
                            shared_variables.pmaster.multicast_group,
                            shared_variables.pmaster.audio_port,
                        ),
                    )
                    count = count + leftover
                # logger.info(f"Sent this many bytes: {count/2}, out of {length/2}")

            elif audio_format == "alaw":
                for i in range(length // block_size):
                    # elapsed_time = time.perf_counter() - self._start_time  // i
                    shared_variables.pmaster.audio_socket.sendto(
                        audioop.lin2alaw(audio_bin_buff[i * block_size: (i + 1) * block_size], 2),
                        (
                            shared_variables.pmaster.multicast_group,
                            shared_variables.pmaster.audio_port,
                        ),
                    )
                    count = count + block_size

                    # Burst the first PACKETS_NUMBER packages to accumulate a buffer to read from,
                    # then go with equal intervals
                    if shared_variables.first_fast_packets < constants.PACKETS_NUMBER:
                        time.sleep(0.004)
                        shared_variables.first_fast_packets = shared_variables.first_fast_packets + 1
                    else:
                        time.sleep(delay)

                # If there is data left over from the buffer, send it at the end without delay
                if leftover != 0:
                    shared_variables.pmaster.audio_socket.sendto(
                        audioop.lin2alaw(audio_bin_buff[length - leftover:], 2),
                        (
                            shared_variables.pmaster.multicast_group,
                            shared_variables.pmaster.audio_port,
                        ),
                    )
                    count = count + leftover
                logger.info(f"Sent this many bytes: {count / 2}, out of {length / 2}")
            else:
                logger.warning(
                    f"Audio format not recognized : {audio_format}"
                )
                self._set_response(400)
            m_queue.task_done()
        # logger.info("Finished paging thread")

    def _encode_send_audio(self, audio_bin_buff, audio_format="", block_size=1920, delay=0.001):
        """
        Takes the PCM 16 audio buffer, encodes it to either uLaw or aLaw and sends it over through the socket
        to the selected clients
        Args:
            audio_bin_buff: Binary IO audio buffers
            audio_format: uLaw or aLaw allowed
            block_size: Packet size to be encoded and sent at a time
            delay: Delay in [s] between the different packets

        Returns: Void
        """

        count = 0
        length = len(audio_bin_buff)
        leftover = length % block_size

        logger.info(f"Using block size: {block_size}, and delay: {delay}")
        # logger.info(f"Value of first fast packets variable: {shared_variables.first_fast_packets}")

        if audio_format == "ulaw":
            for i in range(length // block_size):
                # elapsed_time = time.perf_counter() - self._start_time  // i
                shared_variables.pmaster.audio_socket.sendto(
                    audioop.lin2ulaw(audio_bin_buff[i * block_size: (i + 1) * block_size], 2),
                    (
                        shared_variables.pmaster.multicast_group,
                        shared_variables.pmaster.audio_port,
                    ),
                )
                count = count + block_size

                if shared_variables.first_fast_packets < constants.PACKETS_NUMBER:
                    time.sleep(0.002)
                    shared_variables.first_fast_packets += 1
                else:
                    time.sleep(delay)

            if leftover:
                shared_variables.pmaster.audio_socket.sendto(
                    audioop.lin2ulaw(audio_bin_buff[length - leftover:], 2),
                    (
                        shared_variables.pmaster.multicast_group,
                        shared_variables.pmaster.audio_port,
                    ),
                )
                count = count + leftover
            logger.info(f"Sent this many bytes: {count/2}, out of {length/2}")

        elif audio_format == "alaw":
            for i in range(length // block_size):
                shared_variables.pmaster.audio_socket.sendto(
                    audioop.lin2alaw(audio_bin_buff[i * block_size: (i + 1) * block_size], 2),
                    (
                        shared_variables.pmaster.multicast_group,
                        shared_variables.pmaster.audio_port,
                    ),
                )
                count = count + block_size
                time.sleep(delay)

            if leftover:
                shared_variables.pmaster.audio_socket.sendto(
                    audioop.lin2alaw(audio_bin_buff[length - leftover:], 2),
                    (
                        shared_variables.pmaster.multicast_group,
                        shared_variables.pmaster.audio_port,
                    ),
                )
                count = count + leftover
            logger.info(f"Sent this many bytes: {count}, out of {len(audio_bin_buff)}")
        else:
            logger.warning(
                f"Audio format not recognized : {audio_format}"
            )
            self._set_response(400)

    def _check_user(self, secret):
        # Add timer to the user connection in case of bad messages.
        # timer = Timer("_check_user")
        # timer.start()
        if shared_variables.active_user is None:
            shared_variables.active_user = shared_variables.users.get_user_from_token(
                secret
            )
            # timer.stop()
            return True if shared_variables.active_user is not None else False
        elif shared_variables.users.secret(shared_variables.active_user) != secret:
            # timer.stop()
            return False
        # timer.stop()
        return True

    def _start_stop_page_message(self):
        send_page_group_thread = None
        shared_variables.downloading_recorded_message = constants.YES
        content_length = int(self.headers["Content-Length"])
        message = self.rfile.read(content_length)
        # logger.info("Message received: %s", message)
        message = json.loads(message)

        if self._check_user(message["token"]):
            # logger.info(f"User matched: {shared_variables.active_user}")

            data_block_size = 320
            delay = 0.002
            # logger.info("Sample rate: %s", shared_variables.config["sample_rate"])
            if shared_variables.config["sample_rate"] == 8000:
                data_block_size = 912
                delay = float("{0:.4f}".format(data_block_size / 2 / shared_variables.config["sample_rate"]))
                # delay = 0.02
            elif shared_variables.config["sample_rate"] == 24000:
                data_block_size = 964
                delay = float("{0:.4f}".format(data_block_size / 2 / shared_variables.config["sample_rate"]))
                # delay = 0.02
            else:
                logger.warning("Sample rate not supported")

            logger.info("Message type: %s", message["type"])

            if message["type"] == 8 and shared_variables.last_message_type == 3:

                if self.path == "/paging":
                    # Wait until the queue is empty
                    while shared_variables.paging_message_q.empty() is False:
                        logger.warning(f"Waiting for queue to be emptied: {shared_variables.paging_message_q.qsize()}")
                        time.sleep(3)

                    # logger.info(f"Total buffers sent to queue: {shared_variables.paging_message_q_buff_msg}")
                    # 16 depends on the settings from the front end.
                    logger.info(f"Total received data: {shared_variables.paging_message_q_buff_msg*16}kb")

                    # logger.info("Current shared_variables.paging_message_q_thread value: %s",
                    #             shared_variables.paging_message_q_thread)

                    shared_variables.paging_state = constants.IDLE
                    shared_variables.paging_message_q_buff_msg = 0

                    # First join the queue
                    if shared_variables.paging_message_q is not None:
                        shared_variables.paging_message_q.join()
                        # logger.info("--->>> Queue joined() <<<---")
                        shared_variables.paging_message_q = None

                    if shared_variables.config["sample_rate"] == 24000:
                        # Second join the reading thread
                        if shared_variables.paging_message_q_thread is not None:
                            shared_variables.paging_message_q_thread.join()
                            # logger.info("--->>> Thread joined() <<<---")
                            shared_variables.paging_message_q_thread = None
                    else:
                        # TODO: Find out why the join() doesn't work at 8kHz
                        shared_variables.paging_message_q_thread = None

                # logger.warning("Current send_page_group_thread value: %s",
                #                send_page_group_thread)

                if message["groups"] == "":
                    logger.warning("No groups selected")
                else:
                    if send_page_group_thread is not None:
                        send_page_group_thread.join()
                        send_page_group_thread = None
                    clients = shared_variables.groups.get_clients(message["groups"])
                    shared_variables.pmaster.send_end_group_request(clients)
                shared_variables.last_client = None
                shared_variables.last_message_type = 8
                shared_variables.active_user = None
                shared_variables.downloading_recorded_message = constants.YES

                if self.path == "/recording":
                    shared_variables.record_service.exit()

                logger.info("STOP paging")
                self._set_response(204)
            elif message["type"] == 3 and shared_variables.last_message_type == 8:

                # logger.warning("Current send_page_group_thread value: %s",
                #                send_page_group_thread)

                if message["groups"] == "":
                    logger.warning("No groups selected")
                else:
                    clients = shared_variables.groups.get_clients(message["groups"])
                    shared_variables.last_client = clients
                    send_page_group_thread = Thread(
                        target=shared_variables.pmaster.send_page_group_request,
                        kwargs={"clients": clients, "priority": 0},
                    )
                    send_page_group_thread.daemon = True
                    send_page_group_thread.start()

                    if self.path == "/paging":
                        shared_variables.paging_state = constants.BUSY
                        # TODO: Not tested
                        shared_variables.paging_message_q = Queue(constants.Q_SIZE)
                        shared_variables.paging_message_q_thread = Thread(
                            target=self._encode_send_audio_queue,
                            kwargs={"m_queue": shared_variables.paging_message_q,
                                    "audio_format": shared_variables.config["audio_format"],
                                    "block_size": data_block_size,
                                    "delay": delay},
                            daemon=True,
                        )

                shared_variables.last_message_type = 3
                shared_variables.first_fast_packets = 0
                # logger.warning("Current send_page_group_thread value: %s",
                #                send_page_group_thread)
                logger.info("START paging")
                self._set_response(204)

            elif message["type"] == 3:
                logger.warning("Last message not ended properly. Ending it first!")

                logger.warning("Current send_page_group_thread value: %s",
                               send_page_group_thread)

                if shared_variables.last_client:
                    if send_page_group_thread is not None:
                        send_page_group_thread.join()
                        send_page_group_thread = None
                    shared_variables.pmaster.send_end_group_request(
                        shared_variables.last_client
                    )
                if message["groups"] == "":
                    shared_variables.last_client = None
                    logger.warning("No group selected")
                else:
                    clients = shared_variables.groups.get_clients(message["groups"])
                    shared_variables.last_client = clients
                    send_page_group_thread = Thread(
                        target=shared_variables.pmaster.send_page_group_request,
                        kwargs={"clients": clients, "priority": 0},
                    )
                    send_page_group_thread.daemon = True
                    send_page_group_thread.start()

                    if self.path == "/paging":
                        shared_variables.paging_state = constants.BUSY
                        # TODO: Not tested
                        shared_variables.paging_message_q = Queue(constants.Q_SIZE)
                        shared_variables.paging_message_q_thread = Thread(
                            target=self._encode_send_audio_queue,
                            kwargs={"m_queue": shared_variables.paging_message_q,
                                    "audio_format": shared_variables.config["audio_format"],
                                    "block_size": data_block_size,
                                    "delay": delay},
                            daemon=True,
                        )

                shared_variables.last_message_type = 3
                shared_variables.first_fast_packets = 0
                logger.info("START paging")
                self._set_response(204)

            else:
                logger.warning(f"Internal server error, {message}")
                # The paging state is not ended correctly
                self._set_response(500)

        else:
            error_msg = f"Another user is paging: {shared_variables.active_user}"
            new_user = shared_variables.users.get_user_from_token(message["token"])
            logger.info(f"Received request from {new_user}")
            logger.warning(error_msg)
            self._set_response(409)
            self.send_header("Content-type", "application/json")
            self.end_headers()
            self.wfile.write(json.dumps(error_msg).encode())

    def _return_active_clients(self):
        self.send_response(200)
        self.send_header("Content-type", "application/json")
        self.end_headers()
        self.wfile.write(
            json.dumps(
                {"active_clients": list(shared_variables.pmaster.active_devices.keys())}
            ).encode()
        )

    def _return_users(self):
        try:
            self.send_response(200)
            self.send_header("Content-type", "application/json")
            self.end_headers()
            self.wfile.write(shared_variables.users.bytes)
        except:
            self.send_response(500)

    def _return_groups(self):
        try:
            self.send_response(200)
            self.send_header("Content-type", "application/json")
            self.end_headers()
            self.wfile.write(shared_variables.groups.bytes)
        except Exception as e:
            template = "An exception of type {0} occurred. Arguments:\n{1!r}"
            message = template.format(type(e).__name__, e.args)
            logger.warning(message)
            self.send_response(500)

    def _return_audio_files(self):
        try:
            self.send_response(200)
            self.send_header("Content-type", "application/json")
            self.end_headers()
            self.wfile.write(shared_variables.audio_files.bytes)
        except Exception as e:
            template = "An exception of type {0} occurred. Arguments:\n{1!r}"
            message = template.format(type(e).__name__, e.args)
            logger.warning(message)
            self.send_response(500)

    def _extract_token_and_audio(self, message: bytes) -> Tuple[str, bytes]:
        """
        Extracts the token and audio packet.
        The whole array is assumed to be 16 bit Int array. The token is encoded to its corresponding ASCII code.
        """

        try:
            token = "".join(
                [
                    chr(
                        int.from_bytes(
                            message[2 * i : 2 * i + 2], "little", signed="True"
                        )
                    )
                    for i in range(shared_variables.user_token_length)
                ]
            )
            # logger.info(f"The token in audio message is {token}")
        except Exception as e:
            logger.warning(f"Unable to extract the token. Error: {e}")
            token = "-1"
        audio = message[shared_variables.user_token_length*2:]
        # logger.info("Token and audio extracted")
        return token, audio

    def _create_group(self):
        content_length = int(self.headers["Content-Length"])
        message = self.rfile.read(content_length)
        message = json.loads(message)
        # logger.warning("Inside _create_group. Message: %s", message)  # Anton
        try:
            status, reason = shared_variables.groups.add(message)
            if status:
                self.send_response(201)
                self.send_header("Content-type", "application/json")
                self.end_headers()
                self.wfile.write(json.dumps(reason).encode())
            else:
                self._set_response(400)
        except Exception as e:
            logger.warning(f"Create group error: {e}")
            self._set_response(500)

    def _create_audio_file(self):
        # print("Inside _create_audio_file.")  # Anton
        content_length = int(self.headers["Content-Length"])
        message = self.rfile.read(content_length)
        message = json.loads(message)
        logger.info("Message: %s", message)  # Anton
        try:
            status, reason = shared_variables.audio_files.add(message)
            if status:
                self.send_response(201)
                self.send_header("Content-type", "application/json")
                self.end_headers()
                self.wfile.write(json.dumps(reason).encode())
            else:
                self._set_response(400)
        except Exception as e:
            logger.warning(f"Create audio file error: {e}")
            self._set_response(500)

    def _create_user(self):
        content_length = int(self.headers["Content-Length"])
        message = self.rfile.read(content_length)
        message = json.loads(message)
        try:
            status, reason = shared_variables.users.add(message)
            if status:
                self.send_response(201)
                self.send_header("Content-type", "application/json")
                self.end_headers()
                self.wfile.write(json.dumps(reason).encode())
            else:
                self.send_response(400)
                self.send_header("Content-type", "application/json")
                self.end_headers()
                self.wfile.write(json.dumps(reason).encode())
        except Exception as e:
            logger.warning(f"Create user error: {e}")
            self._set_response(500)

    def _update_group(self):
        content_length = int(self.headers["Content-Length"])
        message = self.rfile.read(content_length)
        message = json.loads(message)
        try:
            status, reason = shared_variables.groups.update(message)
            if status:
                self.send_response(200)
                self.send_header("Content-type", "application/json")
                self.end_headers()
                self.wfile.write(json.dumps(reason).encode())
            elif status == -1:
                self._set_response(404)
            else:
                self._set_response(400)

        except Exception as e:
            logger.warning(f"Create group error: {e}")
            self._set_response(500)

    def _update_audio_file(self):
        content_length = int(self.headers["Content-Length"])
        message = self.rfile.read(content_length)
        message = json.loads(message)
        try:
            status, reason = shared_variables.audio_files.update(message)
            if status:
                self.send_response(200)
                self.send_header("Content-type", "application/json")
                self.end_headers()
                self.wfile.write(json.dumps(reason).encode())
            elif status == -1:
                self._set_response(404)
            else:
                self._set_response(400)

        except Exception as e:
            logger.error(f"Update audio file error: {e}")
            self._set_response(500)

    def _update_user(self):
        content_length = int(self.headers["Content-Length"])
        message = self.rfile.read(content_length)
        message = json.loads(message)
        try:
            status, reason = shared_variables.users.update(message)
            if status:
                self.send_response(200)
                self.send_header("Content-type", "application/json")
                self.end_headers()
                self.wfile.write(json.dumps(reason).encode())
            elif status == -1:
                self._set_response(404)
            else:
                self._set_response(400)

        except Exception as e:
            logger.warning(f"Create group error: {e}")
            self._set_response(500)

    def _delete_group(self):
        content_length = int(self.headers["Content-Length"])
        message = self.rfile.read(content_length)
        message = json.loads(message)
        try:
            status, reason = shared_variables.groups.remove(message)
            if status:
                self._set_response(200)
            else:
                self._set_response(404)
        except Exception as e:
            logger.warning(f"Delete group error: {e}")
            self._set_response(500)

    def _delete_audio_file(self):
        content_length = int(self.headers["Content-Length"])
        message = self.rfile.read(content_length)
        message = json.loads(message)

        try:
            status, reason = shared_variables.audio_files.remove(message)
            if status:
                self._set_response(200)
            else:
                self._set_response(404)
        except Exception as e:
            logger.warning(f"Delete audio file error: {e}")
            self._set_response(500)

    def _delete_user(self):
        content_length = int(self.headers["Content-Length"])
        message = self.rfile.read(content_length)
        message = json.loads(message)
        try:
            status, reason = shared_variables.users.remove(message)
            if status:
                self._set_response(200)
            else:
                self._set_response(404)
        except Exception as e:
            logger.warning(f"Delete group error: {e}")
            self._set_response(500)

    def log_message(self, format, *args) -> None:
        pass


# Method for starting and running  the server
def run_server(
    server_class=HTTPServer,
    handler_class=PagingHttpRequestHandler,
    port: int = 8765,
):
    server_address = ("0.0.0.0", port)
    httpd = server_class(server_address, handler_class)
    logger.info("Starting httpd...")
    try:
        httpd.serve_forever()
    except KeyboardInterrupt:
        pass
    httpd.server_close()
    logger.info("Stopping httpd...")
